/*
 * Copyright (c) 1997-2003 Erez Zadok
 * Copyright (c) 2001-2003 Stony Brook University
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: inode.c,v 1.1 2003/02/09 23:09:11 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_unrmfs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "unrmfs.h"


STATIC int
unrmfs_create(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "unrmfs_create");

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
    /* XXX this could potentially return a negative hidden_dentry! */
    if (err)
	goto out_lock;

    err = unrmfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);


 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    fist_checkinode(dir, "post unrmfs_create");
    print_exit_status(err);
    return err;
}


STATIC dentry_t *
unrmfs_lookup(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry;
    const char *name;
    vnode_t *this_vnode;
    dentry_t *this_dir;
    unsigned int namelen;
    char *encoded_name;
    unsigned int encoded_namelen;

    print_entry_location();
    hidden_dir_dentry = unrmfs_hidden_dentry(dentry->d_parent); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    namelen = dentry->d_name.len;

    if (!strncmp(name + namelen - 5, ".unrm", 5)) {
	err = -ENOENT;
	goto out;
    }

    fist_checkinode(dir, "unrmfs_lookup");

    this_vnode = dir;
    this_dir = hidden_dir_dentry;

    fist_print_dentry("unrmfs_lookup IN", dentry);
    fist_print_dentry("unrmfs_lookup: dentry->d_parent IN", dentry->d_parent);
    fist_print_dentry("unrmfs_lookup: hidden_dir_dentry IN", hidden_dir_dentry);
    fist_print_inode("unrmfs_lookup: dir IN", dir);

    if (hidden_dir_dentry->d_inode)
	fist_print_inode("unrmfs_lookup: hidden_dir_dentry->d_inode",
			 hidden_dir_dentry->d_inode);

    /* must initialize dentry operations */
    dentry->d_op = &unrmfs_dops;

    ;

    /* increase refcount of base dentry (lookup_one[_len] will decrement) */
    // THIS IS RIGHT! (don't "fix" it)
    // NO THIS IS WRONG IN 2.3.99-pre6. lookup_one[_len] will NOT decrement
    // dget(hidden_dir_dentry);

    encoded_namelen = unrmfs_encode_filename(name,
					     namelen,
					     &encoded_name,
					     SKIP_DOTS, dir, dir->i_sb);
    /* will allocate a new hidden dentry if needed */
    hidden_dentry = lookup_one_len(encoded_name, hidden_dir_dentry, encoded_namelen - 1);
    kfree(encoded_name);

    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    ;

    /* update parent directory's atime */
    fist_copy_attr_atime(dir, hidden_dir_dentry->d_inode);
    /* link the upper and lower dentries */
    dtopd(dentry) = (struct unrmfs_dentry_info *) kmalloc(sizeof(struct unrmfs_dentry_info), GFP_KERNEL);
    if (!dtopd(dentry)) {
	err = -ENOMEM;
	goto out_dput;
    }
    dtohd(dentry) = hidden_dentry;


    /* lookup is special: it needs to handle negative dentries */
    if (!hidden_dentry->d_inode) {
	d_add(dentry, NULL);
	fist_print_dentry("unrmfs_lookup OUT hidden_dentry", hidden_dentry);
	goto out;
    }

    fist_dprint(6, "lookup \"%s\" -> inode %d\n", name, hidden_dentry->d_inode->i_ino);
    err = unrmfs_interpose(hidden_dentry, dentry, dir->i_sb, 1);
    if (err)
	goto out_free;

    fist_checkinode(dentry->d_inode, "unrmfs_lookup OUT: dentry->d_inode:");
    fist_checkinode(dir, "unrmfs_lookup OUT: dir");

    fist_print_dentry("unrmfs_lookup OUT hidden_dentry", hidden_dentry);
    fist_print_inode("unrmfs_lookup OUT hidden_inode", hidden_dentry->d_inode);


    /* All is well */
    goto out;

 out_free:
    d_drop(dentry);		/* so that our bad dentry will get destroyed */
    kfree(dtopd(dentry));
    dtopd(dentry) = NULL;	/* be safe */

 out_dput:
    dput(hidden_dentry);

 out:
    fist_print_dentry("unrmfs_lookup OUT", dentry);
    print_exit_status(err);
    return ERR_PTR(err);
}


STATIC int
unrmfs_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_old_dentry = unrmfs_hidden_dentry(old_dentry); /* CPW: Moved below print_entry_location */
    hidden_new_dentry = unrmfs_hidden_dentry(new_dentry);

    fist_checkinode(dir, "unrmfs_link-dir");
    fist_checkinode(old_dentry->d_inode, "unrmfs_link-oldinode");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_dir_dentry = lock_parent(hidden_new_dentry);


    err = vfs_link(hidden_old_dentry,
		   hidden_dir_dentry->d_inode,
		   hidden_new_dentry);
    if (err || !hidden_new_dentry->d_inode)
	goto out_lock;

    err = unrmfs_interpose(hidden_new_dentry, new_dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_new_dentry->d_inode);
    /* propagate number of hard-links */
    old_dentry->d_inode->i_nlink = itohi(old_dentry->d_inode)->i_nlink;

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);
    if (!new_dentry->d_inode)
	d_drop(new_dentry);

    print_exit_status(err);
    return err;
}

/* This will actually perform  a vfs_rename from F to F.unrm. */
STATIC int
unrmfs_unlink(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    inode_t *hidden_dir; 
    dentry_t *hidden_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_dir_dentry;
    char *unrm_name;

    print_entry_location();

    hidden_dir = itohi(dir); /* CPW: Moved below print_entry_location */
    hidden_dentry = unrmfs_hidden_dentry(dentry);

    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "unrmfs_unlink-dir");

    /* We do this twice because it is a rename, not an unlink. */
    dget(hidden_dentry);
    hidden_dir_dentry = get_parent(hidden_dentry);
    hidden_dir_dentry = get_parent(hidden_dentry);

    /* First we need to get the dentry of the .unrm file */
    unrm_name = kmalloc(hidden_dentry->d_name.len + 5 /* strlen(unrm). */ + 1, GFP_KERNEL);
    if (!unrm_name) {
	err = -ENOMEM;
	goto out_dput;
    }
    memcpy(unrm_name, hidden_dentry->d_name.name, hidden_dentry->d_name.len);
    memcpy(unrm_name + hidden_dentry->d_name.len, ".unrm\0", 6);

    hidden_new_dentry = lookup_one_len(unrm_name, hidden_dentry->d_parent, strlen(unrm_name));
    kfree(unrm_name);
    if (IS_ERR(hidden_new_dentry)) {
	err = PTR_ERR(hidden_dentry);
	goto out_dput;	
    }

    double_lock(hidden_dir_dentry, hidden_dir_dentry);
    err = vfs_rename(hidden_dir, hidden_dentry, hidden_dir, hidden_new_dentry);
    if (err) {
	goto out_lock;
    }

    fist_copy_attr_all(dir, hidden_dir);

    d_drop(dentry);

 out_lock:
    // double_unlock will dput the new/old parent dentries whose refcnts
    // were incremented via get_parent above.
    double_unlock(hidden_dir_dentry, hidden_dir_dentry);
    dput(hidden_new_dentry);
 out_dput:
    dput(hidden_dentry);

    fist_checkinode(dir, "post unrmfs_unlink-dir");
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    char *encoded_symname;
    unsigned int encoded_symlen;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "unrmfs_symlink-dir");

    dget(hidden_dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    encoded_symlen = unrmfs_encode_filename(symname,
					    strlen(symname),
					    &encoded_symname,
					    DO_DOTS, dir, dir->i_sb);
    err = vfs_symlink(hidden_dir_dentry->d_inode,
		      hidden_dentry,
		      encoded_symname);
    kfree(encoded_symname);
    if (err || !hidden_dentry->d_inode)
	goto out_lock;
    err = unrmfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    fist_checkinode(dir, "post unrmfs_symlink-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "unrmfs_mkdir-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mkdir(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = unrmfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    /* update number of links on parent directory */
    dir->i_nlink = hidden_dir_dentry->d_inode->i_nlink;

    fist_checkinode(dir, "post unrmfs_mkdir-dir");

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_rmdir(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "unrmfs_rmdir-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_rmdir does that */
	d_delete(hidden_dentry);

 out_lock:
    fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
    /* copy the nlink count for our dentry and our parent's dentry */
    dir->i_nlink =  hidden_dir_dentry->d_inode->i_nlink;

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err)
	d_drop(dentry);

    dput(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "unrmfs_mknod-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mknod(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode,
		    dev);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = unrmfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;
    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    fist_checkinode(dir, "post unrmfs_mknod-dir");
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_rename(inode_t *old_dir, dentry_t *old_dentry,
	      inode_t *new_dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_old_dir_dentry;
    dentry_t *hidden_new_dir_dentry;

    print_entry_location();

    hidden_old_dentry = unrmfs_hidden_dentry(old_dentry);/* CPW: Moved below print_entry_location */
    hidden_new_dentry = unrmfs_hidden_dentry(new_dentry);

    fist_checkinode(old_dir, "unrmfs_rename-old_dir");
    fist_checkinode(new_dir, "unrmfs_rename-new_dir");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_old_dir_dentry = get_parent(hidden_old_dentry);
    hidden_new_dir_dentry = get_parent(hidden_new_dentry);
    double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);


    err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
		     hidden_new_dir_dentry->d_inode, hidden_new_dentry);
    if (err)
	goto out_lock;

    fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
    if (new_dir != old_dir)
	fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);

 out_lock:
    // double_unlock will dput the new/old parent dentries whose refcnts
    // were incremented via get_parent above.
    double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);

    fist_checkinode(new_dir, "post unrmfs_rename-new_dir");
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
    int err;
    dentry_t *hidden_dentry;
    char *decoded_name, *hidden_buf;
    mm_segment_t old_fs;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    fist_print_dentry("unrmfs_readlink IN", dentry);

    if (!hidden_dentry->d_inode->i_op ||
	!hidden_dentry->d_inode->i_op->readlink) {
	err = -EINVAL;
	goto out;
    }

    hidden_buf = kmalloc(bufsiz, GFP_KERNEL);
    if (hidden_buf == NULL) {
	printk("Out of memory.\n");
	err = -ENOMEM;
	goto out;
    }

    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry,
						 hidden_buf,
						 bufsiz);
    set_fs(old_fs);
    if (err >= 0) {
	// don't do this, it's not zero-terminated!!!
	//	fist_dprint(7, "READLINK: link \"%s\", length %d\n", hidden_buf, err);
	err = unrmfs_decode_filename(hidden_buf, err,
				     &decoded_name, DO_DOTS, dentry->d_inode, dentry->d_sb);
	if (err > 0) {
	    if (copy_to_user(buf, decoded_name, err))
		err = -EFAULT;
	    kfree(decoded_name);
	}
	fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);
    }

    kfree(hidden_buf);

 out:
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_follow_link(dentry_t *dentry, struct nameidata *nd)
{
    char *buf;
    int len = PAGE_SIZE, err;
    mm_segment_t old_fs;

    print_entry_location();
    //    fist_print_dentry("unrmfs_follow_link dentry IN", dentry);

    buf = kmalloc(len, GFP_KERNEL);
    if (!buf) {
	err = -ENOMEM;
	goto out;
    }

    /* read the symlink, and then we will follow it */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = dentry->d_inode->i_op->readlink(dentry, buf, len);
    set_fs(old_fs);
    if (err < 0)
	goto out_free;

    buf[err] = 0;	// terminate the buffer -- XXX still needed?

    // XXX: FIXME w/ unrmfs_encode_filename()
    /*
     * vfs_follow_link will increase the nd's mnt refcnt
     * we assume that some other VFS code decrements it.
     */
    err = vfs_follow_link(nd, buf);

 out_free:
    kfree(buf);
 out:
#if 0
    if (err < 0) {
	dput(nd->dentry);
	printk("EZK follow_link() mnt_cnt %d\n", atomic_read(&nd->mnt->mnt_count));
	mntput(nd->mnt);
    }
#endif

    print_exit_status(err);
    return err;
}


/*
 * Function to handle truncations extending the size of the file for non-SCA
 * file systems.
 */
STATIC int
unrmfs_truncate(dentry_t *dentry, loff_t new_length)
{
    inode_t *inode = dentry->d_inode;
    file_t fake_file;
    page_t *tmp_page;
    int index;
    unsigned to;
    int err = 0;

    /*
     * if this is a truncate past the end of the file, call unrmfs_fill_zeros
     */
    if (new_length > inode->i_size) {

	memset(&fake_file, 0, sizeof(fake_file));
	fake_file.f_dentry = dentry;
	index = new_length >> PAGE_CACHE_SHIFT;

	tmp_page = unrmfs_get1page(&fake_file, index);
	if (IS_ERR(tmp_page)) {
	    err = PTR_ERR(tmp_page);
	    goto out;
	}

	to = new_length & ~PAGE_CACHE_MASK;

	/*
	 * 'to' is passed to unrmfs_fill_zeros instead of a 'from'
	 *  because unlike the normal case where we fill in zeros
	 *  till 'from', we need to fill it in with zeros till the
	 *  end of the file
	 */
	err = unrmfs_fill_zeros(&fake_file, tmp_page, to);
    }

 out:
    return err;
}


STATIC int
unrmfs_permission(inode_t *inode, int mask)
{
    inode_t *hidden_inode;
    int mode;
    int err;

    print_entry_location();
    hidden_inode = itohi(inode);/* CPW: Moved below print_entry_location */
    mode = inode->i_mode;

    err = vfs_permission(inode, mask);
    if (err)
	goto out;
    err = permission(hidden_inode, mask);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_inode_revalidate(dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    hidden_inode = hidden_dentry->d_inode;
    //    fist_print_dentry("unrmfs_inode_revalidate IN", dentry);

    if (hidden_inode->i_op && hidden_inode->i_op->revalidate) {
	err = hidden_inode->i_op->revalidate(hidden_dentry);
	if (!err)
	    fist_copy_attr_all(dentry->d_inode, hidden_inode);
    }

    //    fist_print_dentry("unrmfs_inode_revalidate OUT", dentry);
    print_exit_status(err);
    return err;
}


STATIC int
unrmfs_setattr(dentry_t *dentry, struct iattr *ia)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *inode;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = unrmfs_hidden_dentry(dentry);
    inode = dentry->d_inode;
    hidden_inode = itohi(inode);
    fist_checkinode(inode, "unrmfs_setattr");

    if (ia->ia_valid & ATTR_SIZE) {
	err = unrmfs_truncate(dentry, ia->ia_size);
	if (err < 0)
	    goto out;
    }
    err = notify_change(hidden_dentry, ia);

#if defined(FIST_FILTER_DATA) || defined(FIST_FILTER_SCA)
 out:
#endif /* FIST_FILTER_DATA || FIST_FILTER_SCA */
    /*
     * The lower file system might has changed the attributes, even if
     * notify_change above resulted in an error(!)  so we copy the
     * hidden_inode's attributes (and a few more) to our inode.
     */
    fist_copy_attr_all(inode, hidden_inode);

    fist_checkinode(inode, "post unrmfs_setattr");
    print_exit_status(err);
    return err;
}


#if NOT_USED_YET
STATIC int
unrmfs_getattr(dentry_t *dentry, struct iattr *ia)
{
    return -ENOSYS;
}
#endif /* NOT_USED_YET */


struct inode_operations unrmfs_symlink_iops =
{
    readlink:	unrmfs_readlink,
    follow_link: unrmfs_follow_link,
    permission:	unrmfs_permission,
    revalidate:	unrmfs_inode_revalidate,
    setattr:	unrmfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	unrmfs_getattr,
#endif
};

struct inode_operations unrmfs_dir_iops =
{
    create:	unrmfs_create,
    lookup:	unrmfs_lookup,
    link:	unrmfs_link,
    unlink:	unrmfs_unlink,
    symlink:	unrmfs_symlink,
    mkdir:	unrmfs_mkdir,
    rmdir:	unrmfs_rmdir,
    mknod:	unrmfs_mknod,
    rename:	unrmfs_rename,
    /* no readlink/follow_link for non-symlinks */
    // off because we have setattr
    //    truncate:	unrmfs_truncate,
    permission:	unrmfs_permission,
    revalidate:	unrmfs_inode_revalidate,
    setattr:	unrmfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	unrmfs_getattr,
#endif
};

struct inode_operations unrmfs_main_iops =
{
    permission:	unrmfs_permission,
    revalidate:	unrmfs_inode_revalidate,
    setattr:	unrmfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	unrmfs_getattr,
#endif
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
